/* acia.c - 6502 microcomputer emulator
 *
 * Copyright (C) 2017 Timothy Richard Musson
 *
 * Email: Tim Musson <trmusson@gmail.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>
#include <assert.h>

#include "xlib/xutil.h"
#include "xlib/xmem.h"

#include "config.h"
#include "profile.h"
#include "memory.h"
#include "acia.h"


#ifdef DEBUG
#define DEBUG_ACIA_READ     0
#define DEBUG_ACIA_WRITE    0
#endif


#define STATUS_RX           0x02
#define STATUS_TX           0x01
#define CONTROL_RESET       0x03


static Byte acia_read (ACIA *acia, Word offset);
static void acia_write (ACIA *acia, Word offset, Byte b);


extern Environment environment; /* profile.c */


struct _ACIA {

	/* Memory */

	MemoryType     type;

	int            base;
	int            size;
	xbool_t        read_only;

	MemoryFreeData_f  free_data;
	MemoryStoreByte_f store_byte;
	MemoryReadByte_f  read_byte;

	Byte           *data;

	/* ACIA */

	char           *tape_data;
	Byte           status;

};


ACIA*
acia_new (Machine *m, Word base, char *data)
{
	ACIA *acia;

	acia = xnew0 (ACIA, 1);
	mem_init ((Memory*)acia, base, 256, FALSE, MEM_TYPE_ACIA);

	mem_set_read_func ((Memory*)acia, (MemoryReadByte_f)acia_read);
	mem_set_store_func ((Memory*)acia, (MemoryStoreByte_f)acia_write);

	acia->tape_data = data;
	acia->status = STATUS_TX;

	return acia;
}


void
acia_free (ACIA *acia)
{
	if (acia == NULL) {
		return;
	}

	acia->free_data ((Memory*)acia);
	xfree (acia);
}


static Byte
acia_read (ACIA *acia, Word offset)
{

	/* XXX */

	static int pos = 0;
	Byte b;

	/* assert (acia->tape_data != NULL); */

	if ((offset & 1) == 0) {
		b = acia->status;
	}
	else {

		b = acia->tape_data[pos];

		if (pos < 64 * 1024) {
#if DEBUG_ACIA_READ
			fprintf (stderr, "%c", isgraph (b) || isspace (b) ? b : ' ');
#endif
			pos++;
		}

		if (offset == 1) {
			acia->status = acia->status & ~STATUS_RX;
		}

	}


	return b;
}


static void
acia_write (ACIA *acia, Word offset, Byte b)
{
	/* XXX */

#if DEBUG_ACIA_WRITE
	fprintf (stderr, "acia_write...\n");
#endif

	if ((offset & 1) == 0) {
		if (b == CONTROL_RESET) {
			acia->status = STATUS_TX;
		}
	}
	else {

		/* XXX: send b */
		fprintf (stdout, "%c", b);

		acia->status = acia->status & ~STATUS_TX;
	}

#if DEBUG_ACIA_WRITE
	fprintf (stderr, "acia_write: %02X\n", b);
#endif

}


