/* io.c - 6502 microcomputer emulator
 *
 * Copyright (C) 2017 Timothy Richard Musson
 *
 * Email: Tim Musson <trmusson@gmail.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <assert.h>

#include "xlib/xutil.h"
#include "xlib/xmem.h"

#include "config.h"
#include "profile.h"
#include "memory.h"
#include "video.h"
#include "io.h"


#ifdef DEBUG
#define DEBUG_IO     0
#endif


static void io_putc (IO *io, Word offset, Byte b);
static Byte io_getc (IO *io, Word offset);
static Byte io_random (IO *io, Word offset);


extern Environment environment; /* profile.c */


struct _IO {

	/* Memory */

	MemoryType     type;

	int            base;
	int            size;
	xbool_t         read_only;

	MemoryFreeData_f  free_data;
	MemoryStoreByte_f store_byte;
	MemoryReadByte_f  read_byte;

	Byte           *data;

	/* IO */

};


IO*
io_new (Profile *p, Word base, Word size)
{
	IO *io;

	io = xnew0 (IO, 1);

	mem_init ((Memory*)io, base, size, FALSE, MEM_TYPE_IO);

	if (base == p->getc) {
		mem_set_read_func ((Memory*)io, (MemoryReadByte_f)io_getc);
	}
	else if (base == p->putc) {
		mem_set_store_func ((Memory*)io, (MemoryStoreByte_f)io_putc);
	}
	else if (base == p->random) {
		mem_set_read_func ((Memory*)io, (MemoryReadByte_f)io_random);
	}
	else {
		fprintf (stderr, "%s: io_new: Unhandled device address (%d)\n", p->progname, base);
		io_free (io);
		io = NULL;
	}

	return io;
}


void
io_free (IO *io)
{
	if (io == NULL) {
		return;
	}

	io->free_data ((Memory*)io);
	xfree (io);
}


static void
io_putc (IO *io, Word offset, Byte b)
{
	if (environment == ENVIRONMENT_STDIO) {
		fputc (b, stdout);
	}
}


static Byte
io_getc (IO *io, Word offset)
{
	if (environment == ENVIRONMENT_STDIO) {
		return fgetc (stdin);
	}

	return 0;
}


static Byte
io_random (IO *io, Word offset)
{
	return XRAND (256); /* 0 .. 256-1 */
}


