/* memory.c - 6502 microcomputer emulator
 *
 * Copyright (C) 2017 Timothy Richard Musson
 *
 * Email: Tim Musson <trmusson@gmail.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <assert.h>

#include "xlib/xutil.h"
#include "xlib/xmem.h"

#include "memory.h"


#ifdef DEBUG
static void   mem_debug (Memory *m);
void          mem_set_quiet (xbool_t quiet);
#endif
static void   mem_free_data (Memory *m);


#ifdef DEBUG
xbool_t mem_quiet = FALSE;
#endif


void
mem_init (Memory *m, int base, int size, xbool_t randomize, MemoryType type)
{
	int i;

	m->type = type;
	m->base = base;
	m->size = size;

#ifdef DEBUG
	if (!mem_quiet && base < 0) {
		fprintf (stderr, "mem_init: base < 0 (off-bus storage)\n");
	}
#endif

	if (size > 0) {
		m->data = xnew0 (Byte, size);
		if (randomize) {
			for (i = 0; i < size; i++) {
				m->data[i] = XRAND (256);
			}
		}
	}
	else {
		m->data = NULL;
	}

	m->free_data  = (MemoryFreeData_f) mem_free_data;
	m->read_byte  = (MemoryReadByte_f) mem_read_byte;
	m->store_byte = (MemoryStoreByte_f) mem_store_byte;

	m->read_only  = MEM_READ_WRITE;

#ifdef DEBUG
	if (!mem_quiet) {
		mem_debug (m);
	}
#endif
}


Memory*
mem_new (int base, int size, MemoryType type)
{
	Memory *m;

	m = xnew0 (Memory, 1);
	mem_init (m, base, size, FALSE, type);

	return m;
}


static void
mem_free_data (Memory *m)
{
	xfree (m->data);
}


void
mem_free (Memory *m)
{
	mem_free_data (m);
	xfree (m);
}


Byte
mem_read_byte (Memory *m, int offset)
{
	if (offset < m->size) {
		return m->data[offset];
	}

	return 0;
}


void
mem_store_byte (Memory *m, int offset, Byte b)
{
	if (offset < m->size && !m->read_only) {
		m->data[offset] = b;
	}
}


void
mem_set_free_data_func (Memory *m, MemoryFreeData_f func)
{
	m->free_data = func;
}


void
mem_set_read_func (Memory *m, MemoryReadByte_f func)
{
	m->read_byte = func;
}


void
mem_set_store_func (Memory *m, MemoryStoreByte_f func)
{
	m->store_byte = func;
}


MemoryReadByte_f
mem_get_read_func (Memory *m)
{
	return m->read_byte;
}


MemoryStoreByte_f
mem_get_store_func (Memory *m)
{
	return m->store_byte;
}


#ifdef DEBUG
void
mem_set_quiet (xbool_t quiet)
{
	mem_quiet = quiet;
}


static void
mem_debug (Memory *m)
{
	fprintf (stderr, "Allocated %d bytes of ", m->size);
	switch (m->type) {
		case MEM_TYPE_RAM: fprintf (stderr, "RAM"); break;
		case MEM_TYPE_ROM: fprintf (stderr, "ROM"); break;
		case MEM_TYPE_VIDEO: fprintf (stderr, "video memory"); break;
		case MEM_TYPE_KEYBOARD: fprintf (stderr, "keyboard memory"); break;
		case MEM_TYPE_IO: fprintf (stderr, "I/O port memory"); break;
		case MEM_TYPE_ACIA: fprintf (stderr, "ACIA port memory"); break;
		default: fprintf (stderr, "goodness knows what"); break;
	}
	if (m->base == -1) {
		fprintf (stderr, " (off-bus)\n");
	}
	else if (m->size) {
		fprintf (stderr, " (%04X-%04X)\n", m->base, m->base + m->size - 1);
	}
	else {
		fputc ('\n', stderr);
	}
}
#endif

