/* support.c - 6502 microcomputer emulator
 *
 * Copyright (C) 2017 Timothy Richard Musson
 *
 * Email: Tim Musson <trmusson@gmail.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *
 */

#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <err.h>
#include <errno.h>
#include <assert.h>

#include "xlib/xutil.h"
#include "xlib/xmem.h"
#include "xlib/xstr.h"
#include "xlib/xfile.h"

#include "memory.h"
#include "profile.h"
#include "cpu.h"
#include "video.h"
#include "support.h"



int
parse_ascii_integer (const char *s)
{
	char tmp[6];

	if (strncasecmp (s, "0x", 2) == 0) {
		s++;
	}

	if (*s == '$' || *s == 'x' || *s == 'X') {
		s++;
		sscanf (s, "%4[0-9A-Fa-f]", tmp);
		return strtol (tmp, NULL, 16);
	}

	return atoi (s);
}


xbool_t
parse_ascii_bool (const char *s)
{
	assert (s != NULL);

	if (strcasecmp (s, "TRUE") == 0 || strcmp (s, "1") == 0) {
		return TRUE;
	}
	return FALSE;
}


xbool_t
str_contains_char (const char *s, char c)
{
	for (; *s != '\0'; s++) {
		if (*s == c) {
			return TRUE;
		}
	}

	return FALSE;
}


xbool_t
str_valid (const char *s, const char *legal_chars)
{
	const char *start = s;

	assert (s != NULL);

	while (str_contains_char (legal_chars, *s)) {
		s++;
	}

	if (s == start || *s != '\0') {
		return FALSE;
	}

	return TRUE;
}


xbool_t
parse_cpu_type (const char *s, int *type)
{
	if (s == NULL) {
		return FALSE;
	}

	if (strcmp (s, "6502") == 0) {
		*type = CPU_TYPE_6502;
	}
	else {
		return FALSE;
	}
	return TRUE;
}


xbool_t
parse_environment (const char *s, int *type)
{
	if (s == NULL) {
		return FALSE;
	}

	if (strcasecmp (s, "STDIO") == 0) {
		*type = ENVIRONMENT_STDIO;
	}
	else if (strcasecmp (s, "NCURSES") == 0) {
		*type = ENVIRONMENT_NCURSES;
	}
	else if (strcasecmp (s, "SDL") == 0) {
		*type = ENVIRONMENT_SDL;
	}
	else {
		return FALSE;
	}
	return TRUE;
}


xbool_t
parse_keyboard_type (const char *s, int *type)
{
	if (s == NULL) {
		return FALSE;
	}

	if (strcasecmp (s, "STDIO") == 0) {
		*type = 0;
	}
	else if (strcasecmp (s, "NCURSES") == 0) {
		*type = 1;
	}
	else if (strcasecmp (s, "OSI542") == 0) {
		*type = 2;
	}
	else if (strcasecmp (s, "OSI600") == 0) {
		*type = 3;
	}
	else if (strcasecmp (s, "UK101") == 0) {
		*type = 4;
	}
	else if (strcasecmp (s, "UK101_US") == 0) {
		*type = 5;
	}
	else {
		return FALSE;
	}
	return TRUE;
}


xbool_t
parse_video_rgb (const char *s, Byte *red, Byte *green, Byte *blue)
{
	char **split;
	int n;

	/* XXX: deal with additional formats: 0x000, 0x000000, #00, #000000 */
	/* XXX: add error checking */

	split = xstr_split (s, ", ");
	for (n = 0; split[n] != NULL; n++);

	if (n != 3) {
		xstr_freev (split);
		return FALSE;
	}

	*red   = parse_ascii_integer (split[0]);
	*green = parse_ascii_integer (split[1]);
	*blue  = parse_ascii_integer (split[2]);

#if 0
	fprintf (stderr, "RGB: %02X %02X %02X\n", *red, *green, *blue);
#endif

	xstr_freev (split);

	return TRUE;
}


/* Load an ASCII hex file (e.g. D5 F0 A9 42 C0 FF ...) into the
 * pre-allocated storage space, returning the number of bytes read
 * if successful, otherwise -1.
 */

int
load_hex (const char *progname, const char *fname, Memory *memory, int offset)
{
    size_t bufsize;
    ssize_t result;
	char *buf, **split;
	FILE *f;
	int i, count;
	xbool_t okay;
	Byte *cell;

	if ((f = fopen (fname, "r")) == NULL) {
		fprintf (stderr, "%s: Failed to open '%s' (%s)\n",
		    progname, fname, strerror (errno));
		return -1;
    }

#if 0
	fprintf (stderr, "%s\n", fname);
#endif

	cell = memory->data + offset;
    buf = NULL;
	count = 0;
	okay = TRUE;

    while (okay && (result = xfile_getline (&buf, &bufsize, f)) > 0) {

		split = xstr_split (buf, " \n");

		for (i = 0; split[i] != NULL; i++) {

			if (strlen (split[i]) != 2 || !str_valid (split[i], DIGITS_HEX)) {
				fprintf (stderr, "%s: Unexpected value in hex file '%s'\n",
				    progname, fname);
				okay = FALSE;
				break;
			}

			*cell = strtol (split[i], NULL, 16);
#if 0
			fprintf (stderr, "%02X\n", *cell);
#endif
			cell++;
			count++;

			if (count + offset > memory->size) {
				fprintf (stderr, "%s: More data than allocated space (%s)\n",
				    progname, fname);
				okay = FALSE;
				break;
			}

		}
		xstr_freev (split);

	}

    xfree (buf);
    fclose (f);

	return okay ? count : -1;
}


