/* video.c - 6502 microcomputer emulator
 *
 * Copyright (C) 2020 Timothy Richard Musson
 *
 * Email: Tim Musson <trmusson@gmail.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>

#include "xlib/xutil.h"
#include "xlib/xmem.h"

#include "config.h"
#include "memory.h"
#include "rom.h"
#include "profile.h"
#include "video.h"
#include "keyboard.h"
#include "machine.h"


#ifdef ENABLE_SDL
#include <SDL2/SDL.h>
#endif

#ifdef ENABLE_NCURSES
#include <ncurses.h>
#endif


extern Machine *machine;        /* emu.c */
extern Environment environment; /* profile.c */


struct _Video {

	/* Memory */

	MemoryType     type;

	int            base;
	int            size;
	xbool_t         read_only;

	MemoryFreeData_f  free_data;
	MemoryStoreByte_f store_byte;
	MemoryReadByte_f  read_byte;

	Byte           *data;

	/* Video */

	Environment    environment;

	int            width;       /* in characters */
	int            height;      /* in characters */

	Byte           fg[3];      /* foreground colour RGB */
	Byte           bg[3];      /* background colour RGB */

#ifdef ENABLE_SDL

	SDL_Window     *window;
	SDL_Renderer   *renderer;
	SDL_Surface    *surface;
	SDL_Texture    *texture;

	int            screen_width;    /* in pixels */
	int            screen_height;   /* in pixels */

	int            scale_w;
	int            scale_h;

	ROM            *charset;
	int            char_width;      /* bytes across   */
	int            char_height;     /* bytes down     */
	int            char_bytes;      /* bytes per char */
	int            char_count;

	xbool_t         fullscreen;

#endif

#ifdef ENABLE_NCURSES

	WINDOW         *nc_window;

#endif

};


static void video_update (Video *v, Word offset, Byte b);
static void video_store_byte (Video *v, Word offset, Byte b);

#ifdef ENABLE_SDL
static void video_render_character_sdl (Video *v, int x, int y, Byte ch);
#endif


Video*
video_new (Profile *p, xbool_t fullscreen)
{
	Video *v;

	v = xnew0 (Video, 1);

	mem_init ((Memory*)v, p->video_base, p->video_size, !p->ram_clear,
	          MEM_TYPE_VIDEO);

	v->environment = p->environment;
	v->width = p->screen_width;
	v->height = p->screen_height;

	v->fg[0] = p->video_fg[0];
	v->fg[1] = p->video_fg[1];
	v->fg[2] = p->video_fg[2];

	v->bg[0] = p->video_bg[0];
	v->bg[1] = p->video_bg[1];
	v->bg[2] = p->video_bg[2];

	mem_set_store_func ((Memory*)v, (MemoryStoreByte_f)video_store_byte);

#ifdef ENABLE_SDL


	if (v->environment == ENVIRONMENT_SDL) {

		/* Uint32 rmask, gmask, bmask, amask; */
		Uint32 window_flags;
		int charset_bytes;
		int x, y, offset;

		if (p->video_charset_fname == NULL) {
			fprintf (stderr, "%s: No video character-set specified\n",
			         p->progname);
			video_free (v);
			return NULL;
		}

		if ((v->charset = rom_new (p, p->video_charset_fname)) == NULL) {
			video_free (v);
			return NULL;
		}

		v->scale_w = p->video_scale_w;
		v->scale_h = p->video_scale_h;

		v->char_width = p->video_char_width;
		v->char_height = p->video_char_height;
		v->char_count = p->video_char_count;

		v->char_bytes = v->char_width * v->char_height;
		charset_bytes = v->char_bytes * v->char_count;

		if (charset_bytes != ((Memory*)v->charset)->size) {
			fprintf (stderr,
			         "%s: Character set ROM (%d bytes) inconsistent with %d %d x %d characters (%d bytes) (check values in profile '%s')\n",
			         p->progname, charset_bytes, v->char_count,
			         v->char_width, v->char_height,
			         ((Memory*)v->charset)->size, p->fname);
			video_free (v);
			return NULL;
		}

		/* XXX: 8-pixel char width, better if not hard-wired */
		v->screen_width = 8 * v->char_width * v->width;
		v->screen_height = v->char_height * v->height;

		if (SDL_Init (SDL_INIT_VIDEO) < 0) {
			fprintf (stderr, "%s: Couldn't start SDL: %s\n",
			         p->progname, SDL_GetError ());
			video_free (v);
			return NULL;
		}

		v->fullscreen = fullscreen;
		window_flags = SDL_RENDERER_ACCELERATED;
		if (v->fullscreen) {
			window_flags |= SDL_WINDOW_FULLSCREEN;
		}

		if (SDL_CreateWindowAndRenderer (v->screen_width * v->scale_w,
		                                 v->screen_height * v->scale_h,
		                                 window_flags, &v->window,
		                                 &v->renderer) != 0) {
			fprintf (stderr, "%s: Couldn't open SDL window: %s\n",
			         p->progname, SDL_GetError ());
			video_free (v);
			return NULL;
		}

		SDL_SetWindowTitle (v->window, p->simulation_name);
		SDL_RenderSetScale (v->renderer, (float)v->scale_w, (float)v->scale_h);
		SDL_SetRenderDrawColor (v->renderer, v->bg[0], v->bg[1], v->bg[2],
		                        SDL_ALPHA_OPAQUE);
		SDL_RenderClear (v->renderer);

		/* XXX: check result */
		v->surface = SDL_CreateRGBSurface (0, v->screen_width, v->screen_height,
		                                   32, 0, 0, 0, 0);

		/* XXX: check result */
		v->texture = SDL_CreateTexture (v->renderer, SDL_PIXELFORMAT_ARGB8888,
		                                SDL_TEXTUREACCESS_STREAMING,
		                                v->screen_width, v->screen_height);

		/* Draw initial video memory contents, for "authentic" effect */
		for (offset = 0; offset < v->width * v->height; offset++) {
			y = (int)(offset / v->width);
			x = offset - (y * v->width);
			video_render_character_sdl (v, x, y, v->data[offset]);
		}
		SDL_UpdateTexture (v->texture, NULL, v->surface->pixels,
		                   v->screen_width * sizeof (Uint32));
		SDL_RenderClear (v->renderer);
		SDL_RenderCopy (v->renderer, v->texture, NULL, NULL);

		SDL_RenderPresent (v->renderer);

		if (p->reset_timeout > 0) {
			SDL_Delay (p->reset_timeout);
		}
	}

#else

	if (v->environment == ENVIRONMENT_SDL) {
		fprintf (stderr,
		         "%s: SDL support required by this profile is not compiled in\n",
		         p->progname);
		video_free (v);
		return NULL;
	}

#endif

#ifdef ENABLE_NCURSES

	if (v->environment == ENVIRONMENT_NCURSES) {

		initscr ();
		cbreak ();
		noecho ();
		nodelay (stdscr, TRUE);
		start_color ();
		init_pair (1, COLOR_WHITE, COLOR_BLUE); /* XXX: get colours from profile */
		v->nc_window = newwin (v->width, v->height, 1, 1); /* XXX: check if fail */

		nodelay (v->nc_window, TRUE);
		wattron (v->nc_window, COLOR_PAIR (1));
		wclear (v->nc_window);
		curs_set (0);
		wrefresh (v->nc_window);

	}

#else

	if (v->environment == ENVIRONMENT_NCURSES) {
		fprintf (stderr,
		         "%s: NCURSES support required by this profile is not compiled in\n",
		         p->progname);
		video_free (v);
		return NULL;
	}

#endif

	return v;
}


void
video_free (Video *v)
{
	if (v == NULL) {
		return;
	}

	v->free_data ((Memory*)v);

#ifdef ENABLE_SDL

	if (v->environment == ENVIRONMENT_SDL) {

		/* XXX: ought to destroy v->surface? */

		SDL_DestroyTexture (v->texture);
		SDL_DestroyRenderer (v->renderer);
		SDL_DestroyWindow (v->window);
		SDL_Quit ();

		if (v->charset) {
			rom_free (v->charset);
		}

	}

#endif

#ifdef ENABLE_NCURSES

	if (v->environment == ENVIRONMENT_NCURSES) {

		delwin (v->nc_window);
		endwin ();

	}

#endif

	xfree (v);
}


static void
video_store_byte (Video *v, Word offset, Byte b)
{
	if (offset < v->size) {
		mem_store_byte ((Memory*)v, offset, b);
		video_update (v, offset, b);
	}
}


static void
video_update (Video *v, Word offset, Byte b)
{

#ifdef ENABLE_NCURSES

	if (v->environment == ENVIRONMENT_NCURSES) {
		/* XXX */
	}

#endif

#ifdef ENABLE_SDL

	if (v->environment == ENVIRONMENT_SDL) {
		int x, y;
		y = (int)(offset / v->width);
		x = offset - (y * v->width);
		video_render_character_sdl (v, x, y, v->data[offset]);
	}

#endif

}


#ifdef ENABLE_NCURSES
WINDOW*
video_get_window_ncurses (Video *v)
{
	return v->nc_window;
}


static void
video_refresh_ncurses (Video *v)
{
	int row, col, cell;
	Byte b;

	cell = 0;

	for (row = 0; row < v->height; row++) {
		for (col = 0; col < v->width; col++) {
			b = v->data[cell];
			if (!isprint (b)) {
				b = ' ';
			}
			wmove (v->nc_window, row, col);
			waddch (v->nc_window, b);
			cell++;
		}
	}

	wrefresh (v->nc_window);
}
#endif


void
video_set_fullscreen (Video *v, xbool_t fullscreen)
{
#ifdef ENABLE_SDL
	SDL_SetWindowFullscreen (v->window, fullscreen ? SDL_WINDOW_FULLSCREEN : 0);
	v->fullscreen = fullscreen;
#endif
}


void
video_refresh (void)
{
	static int count = 0;

	count++;

	/* XXX: timing here is just a quick hack - needs fixing */
	/* I think SDL has timeout/signal-handling features that might */
	/* make sense for this */

	if (count == 1250) {

		count = 0;

#ifdef ENABLE_SDL
		if (environment == ENVIRONMENT_SDL) {

			Video *v = machine_get_video (machine);
			SDL_UpdateTexture (v->texture, NULL, v->surface->pixels,
			                   v->screen_width * sizeof (Uint32));
			SDL_RenderClear (v->renderer);
			SDL_RenderCopy (v->renderer, v->texture, NULL, NULL);
			SDL_RenderPresent (v->renderer);
		}
#endif

#ifdef ENABLE_NCURSES
		if (environment == ENVIRONMENT_NCURSES) {
			video_refresh_ncurses (machine_get_video (machine));
		}
#endif
	}
}


#ifdef ENABLE_SDL


static void
set_pixel (SDL_Surface *surface, int x, int y, Uint32 color)
{
	Uint32 *pixel;

	/* XXX */
	pixel = (Uint8*)surface->pixels + y * surface->pitch + x * sizeof (*pixel);

	*pixel = color;
}


static void
video_render_character_sdl (Video *v, int x, int y, Byte ch)
{
	int i, j, cell;
	int ch_pixel_width, ch_pixel_height;
	int scr_x, scr_y;
	Byte b;
	Uint32 color;

	ch_pixel_width = v->char_width * 8; /* XXX */
	ch_pixel_height = v->char_height;

	scr_x = x * ch_pixel_width;
	scr_y = y * ch_pixel_height;

	cell = v->char_bytes * ch;

	color = SDL_MapRGB (v->surface->format, v->bg[0], v->bg[1], v->bg[2]);

	for (j = 0; j < ch_pixel_height; j++) {
		for (i = 0; i < ch_pixel_width; i++) {
			set_pixel (v->surface, scr_x + i, scr_y + j, color);
		}
	}

	color = SDL_MapRGB (v->surface->format, v->fg[0], v->fg[1], v->fg[2]);

	for (j = 0; j < ch_pixel_height; j++) {

		b = ((Memory*)v->charset)->data[cell];

		for (i = 0; i < ch_pixel_width; i++) {

			if (TEST_BIT(b, 7 - i)) {
				set_pixel (v->surface, scr_x + i, scr_y + j, color);
			}

		}
		cell++;
	}
}
#endif


void
environment_event_handler (xbool_t *quit)
{

	if (environment == ENVIRONMENT_SDL) {

#ifdef ENABLE_SDL

		SDL_Event sdl_event;
		Keyboard *k;

		k = machine_get_keyboard (machine);

		while (SDL_PollEvent (&sdl_event) != 0) {

			if (sdl_event.type == SDL_QUIT) {
				*quit = TRUE;
				break;
			}
			else if (sdl_event.type == SDL_KEYDOWN) {
				keyboard_event_sdl (k, sdl_event.key.keysym.sym, TRUE);
			}
			else if (sdl_event.type == SDL_KEYUP) {
				keyboard_event_sdl (k, sdl_event.key.keysym.sym, FALSE);
				if (keyboard_get_reset (k)) {
					machine_reset (machine);
				}
				if (keyboard_get_stepping (k)) {
					machine_toggle_stepping (machine);
				}
				if (keyboard_get_step_now (k)) {
					machine_step_now (machine);
				}
				if (keyboard_get_fullscreen_toggle (k)) {
					Video *v = machine_get_video (machine);
					keyboard_set_fullscreen_toggle (k, FALSE);
					video_set_fullscreen (v, !v->fullscreen);
				}
			}

		}
#endif

	}

	if (environment == ENVIRONMENT_NCURSES) {

#ifdef ENABLE_NCURSES

		WINDOW *nc_window;
		int c;

		nc_window = video_get_window_ncurses (machine_get_video (machine));

		c = wgetch (nc_window);
		if (c != -1) {

			keyboard_event_ncurses (machine_get_keyboard (machine), c);

		}

#endif

	}

}


