/* xcsvio.c - CSV/TSV file input/output
 *
 * Copyright (C) 2009, 2014, 2019, 2020 Timothy Richard Musson
 *
 * Email: Tim Musson <trmusson@gmail.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *
 * gcc -D TEST xmem.o xstr.o xstring.o xlist.o xerror.o xcsvio.c -o xcsvio-test
 *
 */

/*

See:

  http://en.wikipedia.org/wiki/Comma-separated_values
  http://www.creativyst.com/Doc/Articles/CSV/CSV01.htm

An example CSV file from Wikipedia:

1997,Ford,E350,"ac, abs, moon",3000.00
1999,Chevy,"Venture ""Extended Edition""","",4900.00
1996,Jeep,Grand Cherokee,"MUST SELL!
air, moon roof, loaded",4799.00

*/


#include <stdlib.h>
#include <stdio.h>
#include <string.h> /* strerror */
#include <ctype.h>  /* isspace */
#include <errno.h>
#include <sys/stat.h>
#include <assert.h>

#include "xutil.h"
#include "xmem.h"
#include "xstr.h"
#include "xstring.h"
#include "xlist.h"
#include "xerror.h"
#include "xcsvio.h"


typedef enum {
	XCSV_TOKEN_NONE = 0,
	XCSV_TOKEN_DATA,
	XCSV_TOKEN_NEW_LINE,
	XCSV_TOKEN_DELIMITER,
	XCSV_TOKEN_BEGIN_QUOTE,
	XCSV_TOKEN_ERROR
} XCsvToken;


struct _XCsv {

	char      *fname;
	FILE      *fp;
	xlist_t   *row;
	xstring_t *data;
	XCsvToken token;
	XCsvToken prev_token;
	xbool_t   quoted;
	char      delimiter;
	xbool_t   is_stdio;

};


/* Open a CSV file for reading.
   The delimiter should usually be either a ',' or '\t' character.
   The err parameter can be NULL.
   Return NULL on failure.  */

XCsv*
xcsv_open_r (const char *fname, char delimiter, XError **err)
{
	FILE *fp;
	XCsv *csv;

	assert (fname != NULL);

	if (strcmp (fname, "-") == 0) {
		fp = stdin;
	}
	else {
		if ((fp = fopen (fname, "r")) == NULL) {
			if (err != NULL) {
				xerror_set (err, 0, strerror (errno));
			}
			return NULL;
		}
	}

	csv = xnew0 (XCsv, 1);

	csv->fp = fp;
	csv->fname = xstr_dup (fname);
	csv->delimiter = delimiter;
	csv->data = xstring_new ("");
	csv->is_stdio = (fp == stdin);

	return csv;
}


/* Open a CSV file for writing.
   The delimiter should usually be either a ',' or '\t' character.
   The err parameter can be NULL.
   Return NULL on failure.  */

XCsv*
xcsv_open_w (const char *fname, char delimiter, XError **err)
{
	FILE *fp;
	XCsv *csv;

	assert (fname != NULL);

	if (strcmp (fname, "-") == 0) {
		fp = stdout;
	}
	else {
		if ((fp = fopen (fname, "w")) == NULL) {
			if (err != NULL) {
				xerror_set (err, 0, strerror (errno));
			}
			return NULL;
		}
	}

	csv = xnew0 (XCsv, 1);

	csv->fp = fp;
	csv->fname = xstr_dup (fname);
	csv->delimiter = delimiter;
	csv->is_stdio = (fp == stdout);

	return csv;
}


/* Close a currently open CSV file.  */

void
xcsv_close (XCsv *csv)
{
	xfree (csv->fname);

	xcsv_row_clear (csv);

	if (csv->data != NULL) {
		xstring_free (csv->data, TRUE);
	}

	if (!csv->is_stdio && csv->fp != NULL) {
		fclose (csv->fp);
	}

	xfree (csv);
}


const char*
xcsv_get_filename (XCsv *csv)
{
	return csv->fname;
}


/* Return a copy of the given string with '"' characters escaped.  */

char*
xcsv_escape_text (const char *text)
{
	xstring_t *gs;
	const char *s;

	assert (text != NULL);

	gs = xstring_new ("");
	for (s = text; *s != '\0'; s++) {
		if (*s == '"') {
			xstring_append_c (gs, '"');
		}
		xstring_append_c (gs, *s);
	}

	return xstring_free (gs, FALSE);
}


/* Free all row data.  */

void
xcsv_row_clear (XCsv *csv)
{
	if (csv->row != NULL) {
		xlist_foreach (csv->row, (XList_f)xfree, NULL);
		xlist_free (&csv->row);
		csv->row = NULL;
	}
}


/* Append a new field to the row, containing the given text.  */

void
xcsv_row_add_field (XCsv *csv, const char *text)
{
	xlist_append (&csv->row, xcsv_escape_text (text));
}


/* Set field n to the given text.  */

void
xcsv_row_set_field (XCsv *csv, int n, const char *text)
{
	xlist_t *target;

	/* Add empty fields prior to n if necessary. */
	while (xlist_length (csv->row) <= n) {
		xcsv_row_add_field (csv, "");
	}

	/* Replace data. */
	target = xlist_nth (csv->row, n);
	xfree (target->data);
	target->data = xstr_dup (text);
}


/* Return the contents of field n.
   Return NULL if undefined or out of bounds.  */

const char*
xcsv_row_get_field (XCsv *csv, int n)
{
	return xlist_nth_data (csv->row, n);
}


/* Copy all fields from another list.  */

void
xcsv_row_copy (XCsv *csv, xlist_t *src)
{
	xlist_t *cur;

	xcsv_row_clear (csv);
	for (cur = src; cur != NULL; cur = cur->next) {
		xcsv_row_add_field (csv, cur->data);
	}
}


/* Write current row to the file.  */

void
xcsv_row_write (XCsv *csv)
{
	xlist_t *cur;

	for (cur = csv->row; cur != NULL; cur = cur->next) {
		fprintf (csv->fp, "\"%s\"", (char*)cur->data);
		if (cur->next != NULL) {
			fprintf (csv->fp, "%c", csv->delimiter);
		}
	}
	fprintf (csv->fp, "\n");
}


static xbool_t
xcsv_find_data (XCsv *csv)
{
	int c;

	if (csv->delimiter != '\t') {
		while (isspace (c = fgetc (csv->fp)) && c != '\n');
	}
	else {
		while ((c = fgetc (csv->fp)) == ' ');
	}

	ungetc (c, csv->fp);

	return c != EOF;
}


static XCsvToken
xcsv_get_token (XCsv *csv)
{
	xbool_t done;
	int c;

	if (csv->prev_token != XCSV_TOKEN_BEGIN_QUOTE) {
		if (!xcsv_find_data (csv)) {
			return XCSV_TOKEN_NONE;
		}
	}

	c = getc (csv->fp);

	if (csv->prev_token == XCSV_TOKEN_NEW_LINE && c == '\n') {
		return XCSV_TOKEN_NEW_LINE;
	}

	if (csv->prev_token == XCSV_TOKEN_DATA) {

		if (c == csv->delimiter) {
			return XCSV_TOKEN_DELIMITER;
		}
		else if (c == '\n') {
			return XCSV_TOKEN_NEW_LINE;
		}

		return XCSV_TOKEN_ERROR;
	}

	if (c == '"' && !csv->quoted) {
		csv->quoted = TRUE;
		return XCSV_TOKEN_BEGIN_QUOTE;
	}

	ungetc (c, csv->fp);
	xstring_assign (csv->data, "");

	while (1) {

		c = getc (csv->fp);

		done = (c == EOF)
			|| (c == csv->delimiter && !csv->quoted)
			|| (c == '\n' && !csv->quoted);

		if (done) {
			ungetc (c, csv->fp);
			return XCSV_TOKEN_DATA;
		}

		if (c == '"') {
			int next = getc (csv->fp);
			if (next != '"') {
				ungetc (next, csv->fp);
				csv->quoted = FALSE;
				return XCSV_TOKEN_DATA;
			}
		}
		xstring_append_c (csv->data, c);
	}
}


/* Read in one row of data.  */

xlist_t*
xcsv_row_read (XCsv *csv, XError **err)
{
	xcsv_row_clear (csv);

	csv->prev_token = XCSV_TOKEN_NONE;

	while (1) {

		csv->token = xcsv_get_token (csv);

		switch (csv->token) {
		case XCSV_TOKEN_DATA:
			xlist_append (&csv->row, xstr_dup ((char*)csv->data->str));
			break;
		case XCSV_TOKEN_NEW_LINE:
		case XCSV_TOKEN_NONE:
			return csv->row;
			break;
		case XCSV_TOKEN_ERROR:
			xerror_set (err, 0, "Error while parsing file");
			return NULL;
			break;
		default:
			break;
		}

		csv->prev_token = csv->token;
	}
}


/* Unlink the row from the XCsv structure and return it to the caller.  */

xlist_t*
xcsv_row_unlink (XCsv *csv)
{
	xlist_t *row;

	row = csv->row;
	csv->row = NULL;

	return row;
}


#ifdef TEST

#define DELIMITER ','  /* ',' for CSV, '\t' for TSV */


/*

A CSV writing example would look something like this:

csv = xcsv_open_w (fname, delimiter, NULL);

for (i = 0; i < 10; i++) {
	xcsv_row_clear (csv);
	xcsv_row_add_field (csv, "foo");
	xcsv_row_add_field (csv, "bar");
	xcsv_row_add_field (csv, "baz");
	xcsv_row_write (csv);
}

xcsv_close (csv);

*/

int
main (int argc, char *argv[])
{
	XError *err = NULL;
	XCsv *csv;
	xlist_t *cur;


	if (argc != 2) {
		fprintf (stderr, "Usage: %s CSVFILE\n", argv[0]);
		exit (EXIT_FAILURE);
	}

	if ((csv = xcsv_open_r (argv[1], DELIMITER, &err)) == NULL) {
		fprintf (stderr, "%s: %s\n", argv[0], err->message);
		xerror_free (err);
		exit (EXIT_FAILURE);
	}

	xlist_pool_init (0);

	while (xcsv_row_read (csv, &err) != NULL) {
		for (cur = csv->row; cur != NULL; cur = cur->next) {
			printf ("{%s}\n", (char*)cur->data);
		}
		printf ("\n");
	}

	if (err) {
		fprintf (stderr, "%s: %s\n", argv[0], err->message);
		xerror_free (err);
		err = NULL;
	}

	xcsv_close (csv);

	xlist_pool_free ();

	return EXIT_SUCCESS;
}


#endif

