/* xfile.c - some basic file handling utilities
 *
 * Copyright (C) 2005, 2020 Timothy Richard Musson
 *
 * Email: Tim Musson <trmusson@gmail.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *
 * gcc -D TEST xmem.o xstr.o xstring.o xfile.c -o xfile-test
 *
 */

#define _GNU_SOURCE /* For getdelim & getline */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <sys/stat.h>
#include <unistd.h>
#include <err.h>
#include <errno.h>

#ifndef XUTIL_H
#include "xutil.h"
#endif

#ifndef XMEM_H
#include "xmem.h"
#endif

#ifndef XSTR_H
#include "xstr.h"
#endif

#ifndef XSTRING_H
#include "xstring.h"
#endif

#ifndef XFILE_H
#include "xfile.h"
#endif

#define XFILE_READ_SIZE        100000
#define XFILE_READ_GROW_SIZE   50000


/* Return a copy of fname, substituting an initial '~' with $HOME if
   appropriate. Return NULL if either: fname is NULL, or unable to 
   to get a value for $HOME.

   Examples:

   "~"      -> "/home/trm"
   "~bad"   -> "~bad"
   "~/good" -> "/home/trm/good"

   Note: this is just a subset of bash tilde-expansion:
   https://www.gnu.org/software/bash/manual/html_node/Tilde-Expansion.html  */

char*
xfile_expand_tilde (const char *fname)
{
	const char *home;
	char *s;
	size_t len_home, len_new;

	if (fname == NULL) {
		return NULL;
	}

	if ((*fname != '~')
	    || (*(fname + 1) != DIR_SEPARATOR_C && *(fname + 1) != '\0')) {
		return xstr_dup (fname);
	}

	if ((home = getenv ("HOME")) == NULL) {
		/* warnx ("getenv failed for HOME"); */
		return NULL;
	}

	len_home = strlen (home);
	len_new = len_home + strlen (fname);

	s = xmalloc (sizeof(char) * len_new);
	strcpy (s, home);
	strcpy (s + len_home, fname + 1);

	return s;
}


/* Strip the last component from a path name.
   The result is a newly allocated string.
   If path is NULL or empty, return ".".
   If path contains no '/' characters, return ".".

   Examples:

   "/usr/lib/"  ->  "/usr"
   "/usr/lib"   ->  "/usr"
   "/usr/"      ->  "/"
   "usr"        ->  "."
   "/"          ->  "/"
   "."          ->  "."
   ".."         ->  "."  */

char*
xfile_dirname (const char *path)
{
	char *s;
	xbool_t ends_with_slash = FALSE;

	if (path == NULL || *path == '\0') {
		return xstr_dup (".");
	}

	s = strrchr (path, DIR_SEPARATOR_C);
	if (s == NULL) {
		return xstr_dup (".");
	}

	ends_with_slash = (*(s + 1) == '\0' ? TRUE : FALSE);

	while (s > path && *s == DIR_SEPARATOR_C) {
		s--;
	}

	if (ends_with_slash) {
		while (s > path && *s != DIR_SEPARATOR_C) {
			s--;
		}
		if (s == path) {
			return xstr_dup (DIR_SEPARATOR_S);
		}
		s--;
	}

	return xstr_ndup (path, s - path + 1);
}


/* Return the last component of a path name.
   The result is a newly allocated string.
   If path is NULL or an empty string, return ".".
   If path contains no '/' characters, return a copy of path.
   If path has a trailing '/', return "".
   This is the same behaviour as the GNU version of basename.

   Examples:

   "/usr/lib/"  ->  ""
   "/usr/lib"   ->  "lib"
   "/usr/"      ->  ""
   "usr"        ->  "usr"
   "/"          ->  ""
   "."          ->  "."
   ".."         ->  ".."  */

char*
xfile_basename (const char *path)
{
	const char *s;

	if (path == NULL || *path == '\0') {
		return xstr_dup (".");
	}
	s = strrchr (path, DIR_SEPARATOR_C);
	if (s == NULL) {
		return xstr_dup (path);
	}
	return xstr_dup (s + 1);
}


/* Return the current working directory as a newly allocated string.  */

char*
xfile_get_cwd (void)
{
#ifdef __USE_GNU

	return getcwd (NULL, 0);

#else

	/* (Example from glibc-doc) */

	size_t size = 100;
	char *buf, *tmp;

	while (TRUE) {
		buf = xmalloc (size);
		tmp = getcwd (buf, size);
		if (tmp == buf) {
			return buf;
		}
		xfree (buf);
		if (errno != ERANGE) {
			return NULL;
		}
		size = size * 2;
	}

#endif
}


/* Return TRUE if the test condition is met for the given file. */

xbool_t
xfile_test (const char *fname, XFileTest test)
{
	struct stat s;

	if ( test & XFILE_TEST_EXISTS ) {
		return access (fname, F_OK) == 0;
	}

	if (stat (fname, &s) == 0) {

		if (test & XFILE_TEST_IS_REGULAR) {
			if (S_ISREG (s.st_mode)) {
				return TRUE;
			}
		}

		if (test & XFILE_TEST_IS_DIR) {
			if (S_ISDIR (s.st_mode)) {
				return TRUE;
			}
		}

	}
	return FALSE;
}


/* Return a string containing the contents of the given FILE.
   If len is non NULL, set len to the number of chars read.  */

char*
xfile_get_contents (FILE* f, size_t *len)
{
	xstring_t *buf;
	int c;

	buf = xstring_alloc (XFILE_READ_SIZE, XFILE_READ_GROW_SIZE);

	while ((c = fgetc (f)) != EOF) {
		xstring_append_c (buf, c);
	}

	if (len != NULL) {
		*len = XSTRING_LEN(buf);
	}

	xstring_append_c (buf, '\0');

	return xstring_free (buf, FALSE);
}


/* Return a string containing the contents of a named file,
   or NULL if unable to open the file.
   If len is non NULL, set len to the number of chars read.  */

char*
xfile_get_contents_with_filename (const char *fname, size_t *len)
{
	FILE *f;
	char *s;

	if ((f = fopen (fname, "r")) == NULL) {
		if (len != NULL) {
			*len = 0;
		}
		return NULL;
	}

	s = xfile_get_contents (f, len);

	fclose (f);

	return s;
}


/* Reads one line from the given FILE and stores it in the
   buffer pointed to by lineptr.
   The buffers grows as needed; the size parameter records this.
   Returns the number of characters read, or -1 if none.
*/

#ifndef __USE_GNU
ssize_t
xfile_getdelim (char **lineptr, size_t *size, int delim, FILE *stream)
{
	/* Fallback for GNU getdelim */

	size_t n = 0;
	int ch = 0;
	char *pos;

	if (lineptr == NULL || size == NULL || stream == NULL) return -1;
	if (*lineptr == NULL) {
		*size = GETLINE_GROW_SIZE;
		*lineptr = xmalloc (*size);
	}

	pos = *lineptr;
	while (ch != delim && (ch = fgetc (stream)) != EOF) {
		if (n == *size) {
			*size += GETLINE_GROW_SIZE;
			*lineptr = xrealloc (*lineptr, *size);
			pos = *lineptr + n;
		}
		*pos = ch;
		pos++;
		n++;
	}
	*pos = '\0';

	if (n > 0) return n;
	return -1;
}


ssize_t
xfile_getline (char **lineptr, size_t *size, FILE *stream)
{
	/* Fallback for GNU getline */
	return xfile_getdelim (lineptr, size, '\n', stream);
}
#endif /* ifndef __USE_GNU */


#ifdef TEST
int
main (int argc, char *argv[])
{
	FILE *f;
	size_t n, len, bufsize;
	ssize_t result;
	const char *fname, *progname;
	char *s, *buf;


	progname = argv[0];

	if (argc != 2 || argv[1][0] == '-') {
		fprintf (stderr, "Usage: %s FILE.txt\n", progname);
		exit (EXIT_FAILURE);
	}

	fname = argv[1];

	s = xfile_expand_tilde ("~");
	printf ("expansion of ~: [%s]\n", s);
	xfree (s);

	s = xfile_expand_tilde ("~foobar");
	printf ("expansion of ~foobar: [%s]\n", s);
	xfree (s);

	s = xfile_expand_tilde ("~/foobar");
	printf ("expansion of ~/foobar: [%s]\n", s);
	xfree (s);

	s = xfile_dirname (fname);
	printf ("dirname: [%s]\n", s);
	xfree (s);

	s = xfile_basename (fname);
	printf ("basename: [%s]\n", s);
	xfree (s);

	s = xfile_basename ("/foo/bar");
	printf ("bar: %s\n", s);
	xfree (s);
	
	s = xfile_basename ("/foo/bar/");
	printf ("bar: %s\n", s);
	xfree (s);

	if (xfile_test (fname, XFILE_TEST_EXISTS)) {
		printf ("file exists\n");
	}
	else {
		printf ("file does not exist\n");
		/* exit (EXIT_SUCCESS); */
	}

	if (xfile_test (fname, XFILE_TEST_IS_REGULAR)) {
		printf ("file is regular\n");
	}
	else {
		if (xfile_test (fname, XFILE_TEST_IS_DIR)) {
			printf ("file is a directory\n");
		}
		else {
			printf ("file is not a directory or a regular file\n");
		}
		/* exit (EXIT_SUCCESS); */
	}

	s = xfile_get_contents_with_filename (fname, &len);
	if (s == NULL) {
		fprintf (stderr, "%s: %s: %s\n", progname, fname, strerror (errno));
		exit (EXIT_FAILURE);
	}

	printf ("%s\n[length: %d]\n", s, (int)len);
	xfree (s);

	f = fopen (fname, "r");
	if (f == NULL) {
		fprintf (stderr, "%s: %s: %s\n", progname, fname, strerror (errno));
		exit (EXIT_FAILURE);
	}

	buf = NULL;

	n = 1;
	len = 0;
	while ((result = xfile_getline (&buf, &bufsize, f)) > 0) {
		printf ("%d: %s", (int)n, buf);
		n++;
		if (result > 0) {
			len += result;
		}
	}
	printf ("[Length: %d]\n", (int)len);

	xfree(buf);

	fclose (f);

	return EXIT_SUCCESS;
}
#endif

