/* xlist.c - singly linked lists
 *
 * Copyright (C) 2005, 2014, 2020 Timothy Richard Musson
 *
 * Email: Tim Musson <trmusson@gmail.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *
 * gcc -D TEST xmem.o xstr.o xlist.c -o xlist-test
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include <err.h>
#include <assert.h>

#ifdef TEST
#include <string.h>
#define DEBUG
#endif

#ifndef XUTIL_H
#include "xutil.h"
#endif

#ifndef XMEM_H
#include "xmem.h"
#endif

#ifndef XSTR_H
#include "xstr.h"
#endif

#ifndef XLIST_H
#include "xlist.h"
#endif



typedef struct _xlist lbuf_t;
typedef struct _lpool lpool_t;


struct _lpool {
	lbuf_t  *buffer;  /* list of node buffers */
	xlist_t *last;    /* last node in current buffer */
	xlist_t *avail;   /* next unused node in current buffer */
	xlist_t *free;    /* list of free nodes in pool */
	size_t  growsz;   /* size allocated for each new node buffer */
	int     count;    /* number of initialized nodes in pool */
	int     taken;    /* nodes currently in use */
};


static lpool_t* lpool_new (size_t growsz);
static void lpool_add_buffer (lpool_t *pool);
static void lpool_free_buffers (lpool_t *pool);


static lpool_t  *pool = NULL;


/* Initialize the list pool.
   Calling this is necessary before the other list functions can be used.  */

void
xlist_pool_init (size_t growsz)
{
	if (growsz < 1) {
		growsz = XLIST_POOL_GROW_SIZE;
	}
	pool = lpool_new (growsz);
}


/* Free the list pool.
   Called at program exit.  */

void
xlist_pool_free (void)
{
	if (xlist_pool_n_taken () != 0) {
		warnx ("** xlist_pool_free: %d nodes lost", xlist_pool_n_taken ());
	}
	lpool_free_buffers (pool);
	xfree (pool);
	pool = NULL;
}


/* Return the number of nodes currently in use.  */

int
xlist_pool_n_taken (void)
{
	return pool->taken;
}


/* Return a new empty node pool.  */

static lpool_t*
lpool_new (size_t growsz)
{
	lpool_t *p;

	assert (growsz > 0);

	p = xnew (lpool_t, 1);
	p->buffer = NULL;
	p->last   = NULL;
	p->avail  = NULL;
	p->free   = NULL;
	p->growsz = growsz;
	p->count  = 0;
	p->taken  = 0;

	return p;
}


/* Grow the node pool.  */

static void
lpool_add_buffer (lpool_t *pool)
{
	lbuf_t *b;

	assert (pool != NULL);

	b = xnew (lbuf_t, 1);
	b->data = xnew (xlist_t, pool->growsz);
	b->next = pool->buffer;
	pool->buffer = b;
	pool->avail = b->data;
	pool->last = (xlist_t*)b->data + pool->growsz;
}


/* Free the buffers and reset the list pool.  */

static void
lpool_free_buffers (lpool_t *pool)
{
	lbuf_t *b, *next;

	assert (pool != NULL);
 
	b = pool->buffer;
	while (b != NULL) {
		next = b->next;
		xfree (b->data);
		xfree (b);
		b = next;
	}
	pool->buffer = NULL;
	pool->last   = NULL;
	pool->avail  = NULL;
	pool->free   = NULL;
	pool->count  = 0;
	pool->taken  = 0;
}


/* Take an unused node from the pool, growing the pool if no
   nodes are available.
   Assign data to the node and return the node to the caller.  */

xlist_t*
xlist_node_get (void *data)
{
	xlist_t *n;
	
	assert (pool != NULL);

	if (pool->free != NULL) {
		n = pool->free;
		pool->free = pool->free->next;
	}
	else {
		if (pool->avail == pool->last) {
			lpool_add_buffer (pool);
		}
		n = pool->avail;
		pool->avail++;
		if (pool->count + 1 < pool->count) {
			warnx ("lnode_get: overflow");
			exit (EXIT_FAILURE);
		}
		pool->count++;
	}

	n->data = data;

	pool->taken++;

	return n;
}


/* Give a node back to the pool.  */

void
xlist_node_put (xlist_t *node)
{
	assert (pool != NULL);
	assert (node != NULL);

	node->next = pool->free;
	pool->free = node;

	pool->taken--;
}


/* Give an entire list back to the pool.  */

void
xlist_free (xlist_t **list)
{
	xlist_t *next;

	assert (list != NULL);

	while (*list != NULL) {
		next = (*list)->next;
		xlist_node_put (*list);
		*list = next;
	}
}


/* Remove the head of a list, give the node back to the pool, and
   return its data to the caller.  */

void*
xlist_pop (xlist_t **list)
{
	xlist_t *n;
	void *data;

	assert (*list != NULL);

	n = *list;
	*list = n->next;
	data = n->data;
	xlist_node_put (n);

	return data;
}


/* Add data to the front of a list.  */

void
xlist_push (xlist_t **list, void *data)
{
	xlist_t *n;

	n = xlist_node_get (data);
	n->next = *list;
	*list = n;
}


/* Add data to the end of a list.  */

void
xlist_append (xlist_t **list, void *data)
{
	xlist_t *n;

	n = xlist_node_get (data);
	n->next = NULL;

	if (*list == NULL) {
		*list = n;
	}
	else {
		xlist_t *cur = *list;
		while (cur->next != NULL) {
			cur = cur->next;
		}
		cur->next = n;
	}
}


/* Transfer the head node of one list to the head of another.  */

void
xlist_move (xlist_t **dest, xlist_t **src)
{
	xlist_t *head;

	assert (*src != NULL);

	head = *src;
	*src = head->next;

	head->next = *dest;
	*dest = head; 
}


/* Append one list to another.  */

void
xlist_join (xlist_t **dest, xlist_t **src)
{
	if (*dest == NULL) {
		*dest = *src;
	}
	else {
		xlist_t *last;
		last = xlist_last (*dest);
		last->next = *src;
	}
	*src = NULL;
}


/* Return a copy of a list.  */

xlist_t*
xlist_copy (xlist_t *list)
{
	xlist_t *new, *tail;

	assert (list != NULL);

	new = NULL;
	xlist_push (&new, list->data);

	tail = new;

	list = list->next;

	while (list != NULL) {
		xlist_push (&tail->next, list->data);
		tail = tail->next;
		list = list->next;
	}
	return new;
}


/* Return the length of a list, or 0 if empty.  */

int
xlist_length (xlist_t *list)
{
	int i = 0;

	while (list != NULL) {
		i++;
		list = list->next;
	}
	return i;
}


/* Remove a node from a list.  */

xlist_t*
xlist_unlink (xlist_t **list, xlist_t *node)
{
	xlist_t *cur, *prev;

	assert (node != NULL);

	cur = *list;

	if (cur == node) {
		*list = node->next;
	}
	else {
		while (cur != NULL) {
			if (cur == node) {
				prev->next = cur->next;
				break;
			}
			prev = cur;
			cur = cur->next;
		}
	}
	return node;
}


/* Remove a node from a list, giving the node back to the pool,
   returning its data to the caller.  */

void*
xlist_remove (xlist_t **list, xlist_t *node)
{
	void *data;

	xlist_unlink (list, node);
	data = node->data;
	xlist_node_put (node);
	return data;
}


/* Insert data at index n of a list.  */

void
xlist_insert_nth (xlist_t **list, int n, void *data)
{
	xlist_t *cur;

	if (n == 0) {
		xlist_push (list, data);
	}
	else {
		cur = *list;
		while (--n > 0 && cur != NULL) {
			cur = cur->next;
		}
		assert (cur != NULL);
		xlist_push (&cur->next, data);
	}
}


/* Insert a new element before node, containing data.  */

void
xlist_insert_before (xlist_t **list, xlist_t *node, void *data)
{
	xlist_t *cur;

	if (*list == NULL || *list == node) {
		xlist_push (list, data);
	}
	else {
		cur = *list;
		while (cur != NULL && cur->next != node) {
			cur = cur->next;
		}
		assert (cur != NULL);
		xlist_push (&cur, data);
	}
}


/* Insert a node into a list according to a comparison function.  */

void
xlist_insert_sorted (xlist_t **list, xlist_t *node, XCmp_f cmp)
{
	xlist_t **cur = list;

	while (*cur != NULL && cmp ((*cur)->data, node->data) < 0) {
		cur = &(*cur)->next;
	}

	node->next = *cur;
	*cur = node;
}


/* Sort a list according to a comparison function.  */

void
xlist_insert_sort (xlist_t **list, XCmp_f cmp)
{
	xlist_t *cur, *next, *tmp = NULL;

	cur = *list;
	while (cur != NULL) {
		next = cur->next;
		xlist_insert_sorted (&tmp, cur, cmp);
		cur = next;
	}

	*list = tmp;
}


/* Return the nth node of a list, or NULL if the list is empty or
   if n is out of range.  */

xlist_t*
xlist_nth (xlist_t *list, int n)
{
	int i = 0;

	if (n < 0) {
		return NULL;
	}

	while (list != NULL && i++ != n) {
		list = list->next;
	}
	return list;
}


/* Return a data pointer from the nth node of a list,
   or NULL if the list is empty or if n is out of range.  */

void*
xlist_nth_data (xlist_t *list, int n)
{
	int i = 0;

	if (n < 0) {
		return NULL;
	}

	while (list != NULL && i++ != n) {
		list = list->next;
	}
	return list == NULL ? NULL : list->data;
}


/* Return the last node of a list.  */

xlist_t*
xlist_last (xlist_t *list)
{
	assert (list != NULL);

	while (list->next != NULL) {
		list = list->next;
	}
	return list;
}


/* Return the first node that points to data,
   or NULL if there's no match.  */

xlist_t*
xlist_find (xlist_t *list, const void *data)
{
	while (list != NULL) {
		if (data == list->data) {
			return list;
		}
		list = list->next;
	}
	return NULL;
}


/* Return the first node that matches data using a comparison function,
   or NULL if there's no match.  */

xlist_t*
xlist_find_cmp (xlist_t *list, const void *data, XCmp_f cmp)
{
	while (list != NULL) {
		if (cmp (data, list->data) == 0) {
			return list;
		}
		list = list->next;
	}
	return NULL;
}


/* Return the position of the node containing the given data,
   or -1 if the node isn't found.  */

int
xlist_index (xlist_t *list, const void *data)
{
	int i = 0;

	while (list != NULL) {
		if (data == list->data) {
			return i;
		}
		list = list->next;
		i++;
	}
	return -1;
}


/* Return the position of the node containing the given data,
   or -1 if the node isn't found, using the given comparison function.  */

int
xlist_index_cmp (xlist_t *list, const void *data, XCmp_f cmp)
{
	int i = 0;

	while (list != NULL) {
		if (cmp (data, list->data) == 0) {
			return i;
		}
		list = list->next;
		i++;
	}
	return -1;
}


/* Return the position of a node in a list,
   or -1 if the node isn't found.  */

int
xlist_index_node (xlist_t *list, xlist_t *node)
{
	int i = 0;

	while (list != NULL) {
		if (list == node) {
			return i;
		}
		list = list->next;
		i++;
	}
	return -1;
}


/* Reverse a list.  */

void
xlist_reverse (xlist_t **list)
{
	xlist_t *tmp;

	tmp = NULL;
	while (*list != NULL) {
		xlist_move (&tmp, list);
	}
	*list = tmp;
}


/* Exchange data between two nodes.  */

void
xlist_swap_data (xlist_t *list, int a, int b)
{
	xlist_t *node_a, *node_b;
	void *tmp;

	if (a != b) {
		node_a = xlist_nth (list, a);
		node_b = xlist_nth (list, b);

		assert (node_a != NULL);
		assert (node_b != NULL);

		tmp = node_a->data;
		node_a->data = node_b->data;
		node_b->data = tmp;
	}
}


/* Rearrange list items into random order.  */

void
xlist_shuffle_data (xlist_t **list)
{
	int i, n;

	n = xlist_length (*list);

	for (i = 0; i < n; i++) {
		xlist_swap_data (*list, i, XRAND(n));
	}
}


/* Remove duplicate elements from a sorted list.  */

void
xlist_uniq (xlist_t *list, XCmp_f cmp, XList_f freefunc)
{
	xlist_t *next;

	if (list == NULL) {
		return;
	}

	while (list->next != NULL) {
		if (cmp (list->data, list->next->data) == 0) {
			next = list->next->next;
			if (freefunc != NULL) {
				freefunc (list->next, NULL);
			}
			xlist_node_put (list->next);
			list->next = next;
		}
		else {
			list = list->next;
		}
	}
}


/* Perform a function on each node of a list.  */

void
xlist_foreach (xlist_t *list, XList_f func, void *data)
{
	while (list != NULL) {
		func (list->data, data);
		list = list->next;
	}
}


#ifdef TEST

static void
my_foreach_func (void *data, void *user_data)
{
	printf ("my_foreach_func: %s\n", (char*)data);
}


static void
showlist (xlist_t *list)
{
	putchar ('(');
	while (list != NULL) {
		printf ("%s", (char*)list->data);
		if (list->next != NULL) {
			printf (", ");
		}
		list = list->next;
	}
	printf (")\n");
}


int
main (int argc, char *argv[])
{
	xlist_t *list, *tmp;

	xlist_pool_init (0);

	list = NULL;
	xlist_push (&list, "4");
	xlist_push (&list, "2");
	xlist_push (&list, "3");
	xlist_push (&list, "5");
	xlist_push (&list, "1");
	showlist (list);

	xlist_reverse (&list);
	showlist (list);

	xlist_insert_sort (&list, (XCmp_f)strcmp);
	showlist (list);

	tmp = xlist_copy (list);
	xlist_join (&list, &tmp);
	showlist (list);
	
	xlist_insert_sort (&list, (XCmp_f)strcmp);
	showlist (list);
	
	xlist_uniq (list, (XCmp_f)strcmp, NULL);
	showlist (list);

	xlist_foreach (list, (XList_f)my_foreach_func, NULL);

	while (!xlist_is_empty (list)) {
		showlist (list);
		xlist_pop (&list);
	}

	xlist_pool_free ();
	return 0;
}

#endif

